/*
*	Bomb Path Picker
*	Copyright (C) 2022 Orinuse
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#pragma semicolon 1
#pragma newdecls required

#include <sourcemod>
#include <sdktools>
#include <profiler>

#define PLUGIN_VERSION "2.2"
#define CONFIG_FILE "data/tf2_bombpathpicker.cfg"

/*========================================================================================
	Changelog:
	
	v2.2 (17-Oct-2022)
	- Upped the config's maximum string character limit from 128 to 255 just for Doppler.
	- Config:
	- - Removed ambiguity in these relays for the following maps:
	- - - Cyberia, Hideout, Powerplant, Waterlogged
	- - - Barren, Downtown, Quetzal, Metro, Underground, Tensai
	- - Added support for Doppler
	
	v2.1 (03-Oct-2022)
	- Fixed string partial-matching not being functional for map names if multiple map names are in use - this prevented maps like Teien or Waterfront from working.
	- Increased the global ArrayStack's cell size as entity names of path choices usually were cutting off too soon.
	- Config:
	- - Resolved ambiguously-named relays for the following maps:
	- - - Dockyard, Outlands, Shiverpeak, Snowpine, Teien
	- - Downpour: Filled in the right path relay names.
	- - Skullcove: Fixed "bompath" typo. Added the missing "s" to "bombpath_holograms_clear_relay"
	- "Random" choice:
	- - Fixed "Random" appearing as a selection on maps have no paths.
	- - Disabled selecting random paths by I/O, as the plugin can't realise what path has been chosen.
	
	v2.0 (05-Sep-2022)
	- Fixed custom inputs also requiring the comma character.
	- Fixed "Random" and "Reset" not cleaning themselves on config reloads.
	- Enhancements to the "Random" selection:
	- - In maps with no entity set for random paths, the plugin uses its own logic.
	- - Paths picked via the selection will use the "(R)" tag for chat messages.
	- Released on AlliedModders.
	
	v1.3 (29-Aug-2022)
	- Implemented custom inputs; Add the pipe character (|) after the relay
	- - E.g. "Reset" "bombpath_navavoid*|Enable"
	- Rewrote the description of the Convars to be less weird to understand.
	
	v1.2 (24-Aug-2022)
	- Fixed the menu being unusable after a map change, which was due to "g_iObjResource" being null.
	- "!bpp" will be denied from being used in non-mvm maps.
	- Translation files are updated with new phrases.
	
	v1.1 (23-Aug-2022)
	- Hotfix for "!bpp" prompting a wave was in-progress during intermissions.
	
	v1.0 (23-Aug-2022)
	- Sent to fingers through DMs as requested. Not released on AM yet because I have no access to my account.

	Special Mentions:
	
*	Meatloaf87 for plugin idea.
*	Fingers for plugin request.

*	Silvers for "[L4D & L4D2] Mission and Weapons - Info Editors" from https://forums.alliedmods.net/showthread.php?t=310586.
*	nosoop for "color_literals.inc" from https://github.com/nosoop/stocksoup/blob/master/color_literals.inc.

*	Jurrell for further understanding of how the standard bombpath I/O chain is setup.
*	MIKObscura for setting up the config file to support a wide set of custom maps by default. Also French translation.
*	royal for his guidance on implementing "g_bCvarAllowMidwave".

======================================================================================*/

ConVar g_hCvarParserDebug, g_hCvarAllowMidwave, g_hMvMMode;
bool g_bCvarParserDebug, g_bCvarAllowMidwave, g_bMvMMode, g_bLateLoad;
int g_iStackPushes, g_iObjResource;
char g_sEntReset[PLATFORM_MAX_PATH];
Menu g_hMenuPaths;
ArrayStack g_hStackPaths;
Profiler g_hProfiler;

////////////////////////////////////////////////////////////////////////////////////////
// Routines on Plugin load
///////////////////////////

// ++ Preload ++ 
// -------------
public Plugin myinfo = 
{
	name = "[TF2] Bomb Path Picker",
	author = "Orinuse",
	description = "Hijack the MvM bomb path to whatever you want.",
	version = PLUGIN_VERSION,
	url = "N/A"
};
public APLRes AskPluginLoad2(Handle myself, bool late, char[] error, int err_max)
{
	EngineVersion engine = GetEngineVersion();
	if( engine != Engine_TF2 )
	{
		strcopy(error, err_max, "Plugin only supports Team Fortress 2.");
		return APLRes_SilentFailure;
	}
	g_bLateLoad = late;
	return APLRes_Success;
}

// ++ Loading ++ 
// -------------
public void OnPluginStart()
{
	LoadTranslations("tf2_bombpathpicker.phrases");

	// Cvars
	CreateConVar("tf2_bpp_version", PLUGIN_VERSION, "'Bomb Path Picker' plugin's version", FCVAR_SPONLY|FCVAR_DONTRECORD|FCVAR_NOTIFY);
	g_hCvarParserDebug = CreateConVar("tf2_bpp_parser_debug", "0", "Prints parser data to the console whenever the config is reloaded.", FCVAR_DONTRECORD);
	g_hCvarAllowMidwave = CreateConVar("tf2_bpp_allow_midwave", "0", "If non-zero, path menu can be accessed during a MvM wave.", FCVAR_NOTIFY);
	g_hMvMMode = FindConVar("tf_gamemode_mvm");
	
	// Commands
	RegConsoleCmd("sm_bpp", 				CmdBPPMenu, "Displays a menu for picking a bomb path.");
	RegAdminCmd("sm_bpp_reload", 			CmdBPPConfigReload, ADMFLAG_ROOT, "Resets and reload the config for 'Bomb Path Picker'.");
	
	g_hCvarParserDebug.AddChangeHook(ConVarChanged_Cvars);
	g_hCvarAllowMidwave.AddChangeHook(ConVarChanged_Cvars);
	g_hMvMMode.AddChangeHook(ConVarChanged_Cvars);
	
	// Menu
	g_hMenuPaths = new Menu(BPPMenuHandler, MENU_ACTIONS_DEFAULT|MenuAction_Display);
	g_hMenuPaths.SetTitle("%T", "Menu Title", LANG_SERVER);
	g_hMenuPaths.ExitButton = false;
	g_hMenuPaths.ExitBackButton = true;
	
	HookEvent("teamplay_round_start", Event_RoundStart, EventHookMode_PostNoCopy);
	
	if ( g_bLateLoad )
	{
		CacheEntities();
	}
	
	AutoExecConfig(true, "tf2_bombpathpicker");
}

void UpdateCvars()
{
	g_bCvarParserDebug = g_hCvarParserDebug.BoolValue;
	g_bCvarAllowMidwave = g_hCvarAllowMidwave.BoolValue;
	g_bMvMMode = g_hMvMMode.BoolValue;
}
public void OnConfigsExecuted()
{
	UpdateCvars();
}
public void ConVarChanged_Cvars(Handle convar, const char[] oldValue, const char[] newValue)
{
	UpdateCvars();
}

void LoadMapPaths()
{
	g_hMenuPaths.RemoveAllItems();
	
	while( g_hStackPaths.Empty == false )
	{
		// 1st push - name it'll use in menu
		// 2nd push - associated logic_relay entity
		char sRelay[PLATFORM_MAX_PATH];
		char sDisplay[32];
		g_hStackPaths.PopString(sRelay, sizeof(sRelay));
		g_hStackPaths.PopString(sDisplay, sizeof(sDisplay));
		
		g_hMenuPaths.AddItem(sRelay, sDisplay);

		if ( g_bCvarParserDebug ) {
			PrintToServer("Stack: Popped %s (%s)", sRelay, sDisplay);
		}
	}

	g_hStackPaths.Clear();
	delete g_hStackPaths;
	
	// Make sure we don't add the random selection to a map that doesn't even have any.. lol
	if( g_hMenuPaths.ItemCount != 0 ) {
		// To magic character, or to not nagic character
	//	g_hMenuPaths.AddItem(g_sEntRandom[0] != 0 ? g_sEntRandom : "&", "Random");
		g_hMenuPaths.AddItem("&", "Random");
	}
}
// Server can start up with plugin running on no maps
// Late loads of plugins will still call this
public void OnMapStart()
{
	LoadConfigSMC();
	LoadMapPaths();
}

////////////////////////////////////////////////////////////////////////////////////////
// ++ Events ++
// Hey, please add something here
/////////////////////////////

void CacheEntities()
{
	g_iObjResource = EntIndexToEntRef(FindEntityByClassname(-1, "tf_objective_resource"));
}
void Event_RoundStart(Event event, const char[] name, bool dontBroadcast)
{
	CacheEntities();
}

////////////////////////////////////////////////////////////////////////////////////////
// ++ Menu ++
// All the menu shenanigans
/////////////////////////////

#define COLOR_CORAL			"\x07FF7F50"
#define COLOR_HONEYDEW		"\x07F0FFF0"
#define COLOR_SUMMERBLUE	"\x0741A3E6" // Signature colour

int BPPMenuHandler(Menu menu, MenuAction action, int client, int index)
{
	if ( action == MenuAction_Display && menu.ItemCount == 0 ) {
		PrintToChat(client, "%t", "Menu No Paths", COLOR_CORAL);
	}

	if( action == MenuAction_Select )
	{
		char sRelay[PLATFORM_MAX_PATH];
		char sDisplay[32];
		int style;
		menu.GetItem(index, sRelay, sizeof(sRelay), style, sDisplay, sizeof(sDisplay));
		
		// Ampersand's the special character for random relay
		if( sRelay[0] == '&' ) {
			// Menu position starts from 0, subtract by 1
			// Avoid the random selection which is located at the bottom, minus another 1.
			SetRandomSeed( GetGameTickCount() );
			int random = GetRandomInt(0,menu.ItemCount - 2);

			menu.GetItem(random, sRelay, sizeof(sRelay), style, sDisplay, sizeof(sDisplay));
			Format(sDisplay, sizeof(sDisplay), "%s (R)", sDisplay);
		}
		
		// Well..... This is it Luigi...
		// I/O time.... IS HERE
		//////////////////////////////////////////////
		int relay = CreateEntityByName("logic_relay");
		if ( !IsValidEntity(relay) ) {
			PrintToChat(client, "%sErr.. something went wrong, try doing the selection again.", COLOR_CORAL);
			return 0;
		}

		// Step 1
		DispatchSpawn(relay);
		ActivateEntity(relay);

		// Step 2 - Clear path logic, then trigger new path logic
		char sOutput[PLATFORM_MAX_PATH];
		DoAddOutput(relay, sOutput, sizeof(sOutput), g_sEntReset, "Trigger", "0.0");
		DoAddOutput(relay, sOutput, sizeof(sOutput), sRelay, strcmp(sDisplay, "Random", true) == 0 ? "PickRandomShuffle" : "Trigger", "1.0");
		
		// Step 3
		AcceptEntityInput(relay, "Trigger", client, client);
		AcceptEntityInput(relay, "Kill", client, client);
		
		// Finisher
		PrintToChatAll("%t", "Menu Select Path", "\x03", sDisplay, COLOR_HONEYDEW);
	}
	return 0;
}
Action CmdBPPMenu(int client, int args)
{
	if ( g_bMvMMode == false ) {
		ReplyToCommand(client, "%t", "Not MvM", COLOR_SUMMERBLUE);
		return Plugin_Handled;
	}

	int iResource = EntRefToEntIndex(g_iObjResource);
	bool bIntermission = view_as<bool>(GetEntProp(iResource, Prop_Send, "m_bMannVsMachineBetweenWaves"));
	if ( g_bCvarAllowMidwave == false && bIntermission == false )
	{
		ReplyToCommand(client, "%t", "Deny Select Path", COLOR_SUMMERBLUE);
	}
	else
	{	
		g_hMenuPaths.Display(client, MENU_TIME_FOREVER);
	}
	
	return Plugin_Handled;
}

#define OUTPUT_FORMAT "OnTrigger %s:%s::%s:1"
void DoAddOutput(int ent, char[] buffer, int size, any ...)
{
	char sArgs[PLATFORM_MAX_PATH];
	char sInput[24];
	char sDelay[6];
	VFormat(sArgs, sizeof(sArgs), "%s", 4);
	VFormat(sInput, sizeof(sInput), "%s", 5);
	VFormat(sDelay, sizeof(sDelay), "%s", 6);

	int comma = StrContains(sArgs, ",");
	if( comma == -1 )
	{
		// Resolve the custom input, and also fix the targetname up
		// to not include anything after the pipe operator.
		/////////////////////////////////////////////////////////////
		int pipe = StrContains(sArgs, "|");
		if ( pipe != -1 ) {
			strcopy(sInput, sizeof(sInput), sArgs[pipe+1]);
			sArgs[pipe] = 0;
		}
		
		Format(buffer, size, OUTPUT_FORMAT, sArgs, sInput, sDelay);
		SetVariantString(buffer);
		AcceptEntityInput(ent, "AddOutput");
	}
	else
	{
		// We have *both* multiple relays and also custom inputs,
		// and we'll deal with the comma-separated entities first
		// before moving on to the pipe character.
		//
		// I somehow replicated Silvers's black magic sorcery out of
		// luck... lol. Thank you Silvers for cursing me
		/////////////////////////////////////////////////////////////
	
		int last = 0;
		int len = strlen(sArgs) + 2;
		char[] newArgs = new char[len];
		StrCat(newArgs, len, sArgs);
		StrCat(newArgs, len, ",");

		while( (comma = StrContains(newArgs[last], ",")) != -1 )
		{
			// Transform the comma into a string terminator
			newArgs[last + comma] = 0;
			
			// Look for the pipe character first.
			int pipe = StrContains(newArgs[last], "|");
			if ( pipe != -1 )
			{
				newArgs[last + pipe] = 0;

				// Now we gotta remember the name separately
				char sName[32];
				char sInputNew[24];
				strcopy(sName, sizeof(sName), newArgs[last]);
				strcopy(sInputNew, sizeof(sInputNew), newArgs[last + pipe+1]);

				newArgs[last + pipe] = '|';
				Format(buffer, size, OUTPUT_FORMAT, sName, sInputNew, sDelay);
			}
			else
			{
				Format(buffer, size, OUTPUT_FORMAT, newArgs[last], sInput, sDelay);
			}

			SetVariantString(buffer);
			AcceptEntityInput(ent, "AddOutput");
			
			// Okay, revert the mess we did or StrContains() fails to work.
			newArgs[last + comma] = ',';
			last += comma + 1;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////
// ++ SMCParser ++
// Configuration file handling
/////////////////////////////

#define MAX_SECTION_DEPTH 2
#define SECTION_ROOT	1
#define SECTION_MAP 	2
#define SECTION_PATHS 	3

bool g_bAllowSectionRead;
int g_iSectionDepth;

void LoadConfigSMC()
{
	char sPath[PLATFORM_MAX_PATH];
	BuildPath(Path_SM, sPath, sizeof(sPath), CONFIG_FILE);

	if( FileExists(sPath) )
	{
		SMCParser parser = new SMCParser();
		g_hStackPaths = new ArrayStack(ByteCountToCells(PLATFORM_MAX_PATH));
		SMC_SetReaders(parser, CfgSMC_EnterSection, CfgSMC_KeyValue, CfgSMC_LeaveSection);
		parser.OnStart = CfgSMC_OnStart;
		parser.OnEnd = CfgSMC_OnEnd;

		// Setup error logging
		char error[128];
		int line, col;
		SMCError result = parser.ParseFile(sPath, line, col);
		if( result != SMCError_Okay )
		{
			if( parser.GetErrorString(result, error, sizeof(error)) )
			{
				SetFailState("SMC CONFIG ERROR: #%d, %s. (line %d, column %d)", result, error, line, col);
			}
			else
			{
				SetFailState("Unable to parse config. Bad format? Check for missing { } etc.");
			}
		}

		delete parser;
		return;
	}
	SetFailState("Could not find the config '%s'!", sPath);
}

void CfgSMC_OnStart(SMCParser parser)
{
	g_sEntReset[0] = 0;
}
SMCResult CfgSMC_EnterSection(SMCParser parser, const char[] section, bool opt_quotes)
{
	g_iSectionDepth++;
	
	if ( g_iSectionDepth == SECTION_MAP )
	{
		char sMap[32];
		GetCurrentMap(sMap, sizeof(sMap));

		if ( g_bCvarParserDebug ) {
			PrintToServer("\n+++++ MAP - %s +++++", section);
			PrintToServer("Depth: %i", g_iSectionDepth);
		}
		
		// We used to split strings by commas in code, but now we don't care, just scan the whole string at once
		if( StrContains(section, sMap, false) != -1 )
		{
			g_bAllowSectionRead = true;
			if ( g_bCvarParserDebug ) {
				PrintToServer("!!!! Matched, Pushing keys to stack !!!!!");
			}
		}
		else
		{
			int comma;
			int last = 0;
			int len = strlen(section) + 2;
			char[] newSection = new char[len];
			StrCat(newSection, len, section);
			StrCat(newSection, len, ",");
		
			while( (comma = StrContains(newSection[last], ",")) != -1 )
			{
				// Transform the comma into a string terminator
				newSection[last + comma] = 0;

				if( StrContains(sMap, newSection[last], false) != -1 ) {
					g_bAllowSectionRead = true;
					if ( g_bCvarParserDebug ) {
						PrintToServer("!!!! Matched, Pushing keys to stack !!!!!");
					}
				}
				
				// Okay, revert the mess we did or StrContains() fails to work.
				newSection[last + comma] = ',';
				last += comma + 1;
			}
		}
	}
	return SMCParse_Continue;
}
SMCResult CfgSMC_KeyValue(SMCParser smc, const char[] key, const char[] value, bool key_quotes, bool value_quotes)
{
	if ( g_bAllowSectionRead == true && g_iSectionDepth == SECTION_PATHS )
	{
		// 1st push - name it'll use in menu
		// 2nd push - associated logic_relay entity
		g_hStackPaths.PushString(key);
		g_hStackPaths.PushString(value);
		g_iStackPushes++;

		if ( g_bCvarParserDebug ) {
			PrintToServer("Stack: Pushed %s (%s)", key, value);
		}
	}
	else if ( g_bAllowSectionRead == true && g_iSectionDepth == SECTION_MAP )
	{
		if ( strcmp(key, "Reset", false) == 0 ) {
			strcopy(g_sEntReset, sizeof(g_sEntReset), value);
		}
		/* else if ( strcmp(key, "Random", false) == 0 ) {
			strcopy(g_sEntRandom, sizeof(g_sEntRandom), value);
		} */
	}
	
	return SMCParse_Continue;
}
SMCResult CfgSMC_LeaveSection(SMCParser parser)
{
	g_iSectionDepth--;
	
	if( g_bAllowSectionRead == true && g_iSectionDepth == SECTION_ROOT )
	{
		if ( g_bCvarParserDebug ) {
			PrintToServer("------ Depth %i ------\n", g_iSectionDepth);
		}
		return SMCParse_Halt;
	}
	return SMCParse_Continue;
}
void CfgSMC_OnEnd(SMCParser parser, bool halted, bool failed)
{
	g_bAllowSectionRead = false;
	g_iSectionDepth = 0;

	if ( failed ) {
		SetFailState("Failed to load %s.", CONFIG_FILE);
	}
	else if ( halted ) {
		if ( g_bCvarParserDebug ) {
			PrintToServer("============");
			PrintToServer("== HALTED ==");
			PrintToServer("============");
			PrintToServer("Depth: %i\n", g_iSectionDepth);
		}
	}
}

Action CmdBPPConfigReload(int client, int args)
{
	if ( g_bMvMMode == false ) {
		ReplyToCommand(client, "%t", "Not MvM", COLOR_SUMMERBLUE);
		return Plugin_Handled;
	}

	g_hProfiler = CreateProfiler();
	g_hProfiler.Start();
	LoadConfigSMC();
	LoadMapPaths();
	g_hProfiler.Stop();
	
	ReplyToCommand(client, "%s[BPP]\x01 Config Reloaded. \x04(took %fs)", COLOR_SUMMERBLUE, g_hProfiler.Time);
	delete g_hProfiler;
	return Plugin_Handled;
}